# Generated by Django 3.1.7 on 2021-10-14 09:04

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0003_add_permission_group"),
    ]

    operations = [
        migrations.AddField(
            model_name="data",
            name="process_resources",
            field=models.JSONField(default=dict),
        ),
    ]
